#!/usr/bin/perl -w

## pic-heterozygosity_v1.1.pl
## Linda Ballard, 21 April 2009
##                USDA-ARS MSA Genomics and Bioinformatics Research Unit
##                linda.ballard at ars.usda.gov
##
##  v1.1:  24 August 2011
##         changed way alleles (TPA) are counted.  Previously was undercounting by 1,2,or 3,
##         causing a slight error in PIC_cross values.
##  Citation:
##		Renee S. Arias, Linda L. Ballard, Brian E. Scheffler
##		UPIC: Perl scripts to determine the number of SSR markers to run
##		Bioinformation 3(8): 352-360, 21 April 2009
##
## Input:  genemapper output after hand processing
## Required Columns 2: Marker name (no spaces)
##                  3: DNA Sample/Line Name (alphanumeric)
##                  4: Amplicon/Peak size (whole number)
## Match input file columns to this, or modify script to match file. 
##          
## Output: file containing PIC values for each marker 
##         and percent heterozygosity across all markers for each line

#use Math::Combinatorics;
#use Data::Dumper; ##testing

#hashes keyed by marker-line, marker, marker-peak
my (%ml_HoA, %m_HoA, %mp_HoA) = ((),(),());

##create hash of hashes that associates heterozyg values with each line 
my %hetero_HoH = (); ## [0]markername [i]count of peaks per line(i)

##create hash containing heterozygosity values for each line
my %hetero_by_line = (); ##key=line_number value=heterozygosity score

##create hash of hashes for upic calculation
my %upic_HoH = (); ##key1=marker, key2=line, value=peak

my @array = (); my @mlines=();
my ($maxlines, $inlinefreq) = (0,1);
my ($c1,$c2,$c3, ) = (1,2,3); #marker, line, peak

printf "Input Requirement:\n\tMARKER in column 2\n\tSAMPLE in column 3\n\tSIZE in column 4\n";
printf "Delimited file of GeneMapper data: ";
chomp ($gmfile=<STDIN>);
open(GMIN, $gmfile) or die "Couldn't open $gmfile: $!\n"; 

$gmfile =~ s/\.\w+$//g;
$pic_hetero_out = $gmfile."\_pic_hetero\.csv";
open(PICOUT, ">$pic_hetero_out") or die "Couldn't create $pic_hetero_out: $!\n"; 

##OPTIONAL: uncomment next 6 lines if input file columns are different from example
#printf "Column number of Markers: ";
#chomp ($c1=<STDIN>);  $c1--; #adjust for array
#printf "Column number of Lines: ";
#chomp ($c2=<STDIN>);  $c2--; #adjust for array
#printf "Column number of Peaks: ";
#chomp ($c3=<STDIN>);  $c3--; #adjust for array

sub indexArray($@)
{
 	my $s=shift;
 	$_ eq $s && return @_ while $_=pop;
 	-1;
}


while (<GMIN>){
	next if 1..1;    ##skip header
	next if /^\s*$/; ##skip blank lines
	##read GMIN into array and split into components
	chomp; @_=split;
	$marker = $_[$c1];
	$line  = $_[$c2];  
	$peak   = $_[$c3];
	
	
	##if lines are named numerically ie 1,2,3, etc.   this works
	#if ($line > $maxlines) { $maxlines=$line; } ##how many lines in input?
	
	##make script accept lines named alpha or numerically
	##NOTE: this is adding a blank entry
	if (!grep /$line/, @mlines) { 
		push @mlines, $line;		
	}
	$line_index = indexArray($line, @mlines) + 1;
	#printf "$marker  $line  $peak  $line_index\n";
	
	$markerline = $marker."_".$line_index;            ##per peak: 1 or more records
	
	##data stored in groups from input
	push @{ $m_HoA{$marker} }, $line_index;
	push @{ $ml_HoA{$markerline} }, $peak;	##hash is unsorted	

	##data stored by row from input, use to calculate inline frequency for PICs
	push @array, ([$marker, $markerline, $line_index, $peak, $inlinefreq, $peakct]); #sorted	
}
close (GMIN);

$maxlines = @mlines;
#printf "maxlines $maxlines  test $testmaxlines\n\n";
#for ($a = 0; $a <= $#mlines; $a++){
#	printf "$mlines[$a] $a ";
#	}

###############################################
##calculate number of alleles with 1 or more peaks per marker
## from hash grouping input by marker

$marker_ct = 0;	
for $marker (keys %m_HoA) {	
	($poss_allele_ct, $count) = (0,0);
	
	for ($h=1; $h<=$maxlines; $h++) { 
		$hetero_HoH{$marker}{$h} = 0; ##load 0 maxline-times to each marker in hash for hetero calc;
	}
	for $i (0 .. $#{ $m_HoA{$marker} } ) { ##loop thru each marker in %m_HoA
		for ($p=1; $p<=$maxlines; $p++) { ##loop thru each line within marker	    
			if ($m_HoA{$marker}[$i] == $p) {			
				##increment line count
				$count = $hetero_HoH{$marker}{$p};
				$count++;
				$hetero_HoH{$marker}{$p} = $count;			
			} 									
		}
		
	}
	##put count of possible alleles in 1st column of %m_HoA
	$poss_allele_ct = @{$m_HoA{$marker}};		
	unshift @{ $m_HoA{$marker} }, $poss_allele_ct;	
	$marker_ct++;
	
}
printf "\n$marker_ct markers in file\n\n";

#############################################
##calculate heterozygosity for each line
##
##  sum of peak freq counts/# of lines >1 peak, across markers per line
##  check peakct sum for each marker:
##   if peakct = 0, no change to num or denom
##   if peakct = 1, denom+1
##   if peakct > 1, num+1 and denom+1

my ($num, $denom) = (0,0);
printf "\nPercent heterozygous loci by Line\n";
printf PICOUT "Line heterozygous_loci\n";

for $h (1 .. $maxlines) {
	$h_line = $mlines[$h - 1];
	printf "$h_line  ";
	#printf PICOUT "$h ";
	for $marker (keys %hetero_HoH) {	
		$peakct = $hetero_HoH{$marker}{$h};		
		if ($peakct == 0) {
			##do nothing (yet) 
		} elsif ($peakct == 1) {
			$denom++;
		} elsif ($peakct > 1 ) {
			$num++; $denom++;
		} else { printf "error at marker $marker line $h\n"; }
	}
	if ($denom!=0){
		#printf "$h ";
		$heterozygosity = $num/$denom;
		printf "$num\/$denom \= $heterozygosity\n";
		$p_hetero = sprintf "%.2f", $heterozygosity * 100;
		printf PICOUT "$h_line $p_hetero\n";
	}
	($num, $denom) = (0,0);
}
printf "\n"; printf PICOUT "\n";
$count = 0;

#############################################
## calculate pic values for each marker
## Step 1: 
##         calculate "within line" frequency	
for $markerline (keys %ml_HoA) {
	$count = $#{ $ml_HoA{$markerline} } + 1; ##how many peaks per line within marker			
	$inlinefreq = 1/$count; #aka psubj
	for $h (0 .. $#array) {
		if ($array[$h][1] eq $markerline) {
			$array[$h][4] = $inlinefreq;
			$array[$h][5] = $count; #for use in pic_cross calculation			
		}
	}		
}
#for $h (0 .. $#array) {
#	if ($array[$h][0] eq "0009_a") {
#			printf	"0009_a $array[$h][2] $array[$h][4]\n";
#	}
#}	

## Step 2: 
##        rearrange data so inline frequency is assigned to each possible marker/peak pair
##        [$marker, $markerline, $line, $peak, $inlinefreq, $peakct]
for ($x=0; $x<=$#array; $x++) {	
	$mark = $array[$x][0];
	$line = $array[$x][2];
	$peak = $array[$x][3];
	$psubj = $array[$x][4]; #aka inlinefreq
	$peakct = $array[$x][5]; 
	$marker_peak = $mark.':'.$peak;
	push @{ $mp_HoA{$marker_peak} }, $array[$x][4];  #inlinefreq to calc F
}	

## Step 3:
##        calculate allele peak frequency -> F
$allele_sum = 0; $dmy = 0;
for $marker_peak (keys %mp_HoA) {
    ($marker1, $dmy) = split /:/, $marker_peak; #splits on first occurence of colon
   	
    $tpa = $m_HoA{$marker1}[0]; ##total possible alleles
	for $i (0 .. $#{ $mp_HoA{$marker_peak}} ) {
		$allele_sum += $mp_HoA{$marker_peak}[$i]; #sum all the inlinefreqs
	}
	#prevent program from stalling out on a div/0 error caused by empty data
	if ($tpa == 0) { $tpa = 0.000000000000000000001; printf "OOPS $marker1 has no alleles\n"; } 
	
	$F = $allele_sum/$tpa;
	push @{ $pic_HoA{$marker1} }, $F;
	$allele_sum = 0;
	#print "$marker1 $tpa $F\n";
}


## Step 4:
##        calculate PIC values (self, equifreq)
my ($F, $Fsquare, $sum_Fsquare, $Fquad, $sum_Fquad,$TwoFsquare, $sum_TwoFsquare) = (0,0,0,0,0,0,0);
printf "\n\nMarker PIC_self PIC_equifreq PIC_cross\n";
printf PICOUT "Marker sum_Fsquare PIC_self PIC_equifreq PIC_cross\n";

for $marker1 (sort keys %pic_HoA) {
	printf "$marker1  ";
	printf PICOUT "$marker1 ";
	$peakct = $#{ $pic_HoA{$marker1} } + 1;  
	$psubj = 1/$peakct;
	$psubj_square = $psubj * $psubj; 

	for $q (0 .. $#{ $pic_HoA{$marker1} } ) {
		$F = $pic_HoA{$marker1}[$q];
		$Fsquare = $F*$F;
		$psubi_psubj = $Fsquare * $psubj_square; #piccross value
		$cross_sum += $psubi_psubj;
		$sum_Fsquare += $Fsquare;
		$TwoFsquare = $Fsquare * 2;
		$sum_TwoFsquare += $TwoFsquare;
		$Fquad = $F*$F*$F*$F;
		$sum_Fquad += $Fquad;
	}
		
	##picself:  self fertilized organisms
	##          1 - (allele_sum/tpa * allele_sum/tpa)
	$picself = 1 - $sum_Fsquare;
	
	##picequi:  equifrequent alleles
	##          $picself - $sum_Fquad
	$picequi = $picself - $sum_Fquad;
	
	##pic_cross: cross fertilized organisms that have equifrequent alleles
	$tpa = $m_HoA{$marker1}[0]; ##total possible alleles
	$cross_mult = $cross_sum * $tpa;
	$piccross = $picself - $cross_mult;
	
	printf PICOUT "$sum_Fsquare $picself $picequi $piccross\n";
	printf "%.5f   ", $picself;
	printf "%.5f   ", $picequi;
	printf "%.5f\n", $piccross;
	($F, $Fsquare, $sum_Fsquare, $Fquad, $sum_Fquad,$TwoFsquare, $sum_TwoFsquare) = (0,0,0,0,0,0,0);
	($psubi_psubj, $psubj_square, $cross_sum, $tpa, $cross_mult) = (0,0,0,0,0);
}

close (PICOUT);

##what's in that hash? To find out, uncomment next 2 lines & USE statement.
#$Data::Dumper::Indent = 1; 
#print Dumper \%m_HoA;

exit;

=pod

Arias Notes

A  Raw data from genemapper
B  Hand processed - removing bad peaks
C  Formula -> PIC & Heterozygosity

1. Generate Data with GeneMapper
2. Mark acceptable peaks with an X (manually modify peak size based on comparison of all peaks per marker, so they bin properly
   (distances less than 1 bp are not separated by rounding up/down).
   
   FORMATTING:
   Insert a Marker and a Line column after column B (SampleFileName).
   File with marker & line data.
   Make sure there is a value in every row of columns A-I.  Missing values will cause data shift.  
    
3. Generate a table of LINES TESTED(header), MARKER(rows), PEAK SIZES (data)
        Note: peak sizes do not need to be rounded up/down - when you save as txt from Excel, it gets done.
4. Sort by size within marker and consolidate to a single row per size 
   with an 1 on LINES that had at least one peak
   0 on LINES with no peak.
5. Generate a within-line frequency
   Replace 1s with result of dividing 1 by NumberOfDetectedPeaksOnALine
6. Sum across lines (col U)
   Generate a count of TotalPossibleAlleles (col V)
   Calculate allele frequency (col W) U/V
   Calculate F2 (col X) W*W
   Sum col X and col W within marker
   Calculate PIC value (1-x)

=cut

exit;